(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON-Web.
 *
 * UMLMON-Web is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON-Web is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON-Web; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: definitions.ml 352 2004-04-07 22:19:37Z gerd $
 * ----------------------------------------------------------------------
 *
 *)

(* This module defines a virtual class "error_behaviour" for two 
 * purposes:
 * - uniform handling of error situations
 * - basic authentication checks that should be done for each click
 *
 * The dialogs (except startpage) should inherit from error_behaviour,
 * and they should define these methods:
 * - try_prepare_page instead of "prepare_page"
 * - try_handle instead of "handle"
 * - check_auth to get the basic auth checks
 *
 *)


open Wd_dialog
open Wd_run_cgi
open Wd_types

open Printf

exception Login_not_permitted


let errid_of_error error =
  match error with
    | Login_not_permitted                  -> "login-not-permitted", ""
    | Common.Wrong_password                -> "wrong-password", ""
    | Common.Remote_error msg              -> "remote-error", msg
    | Common.Interrupted                   -> "interrupted", ""
    | _                                    -> "internal-error", (Printexc.to_string error)
;;


let switch_to_dialog self dlg =
  (* Switch from dialog [self] to [dlg] as substitute for the current
   * dialog
   *)
  dlg # set_variable "nav_prev" (self#variable "nav_prev");
  raise(Change_dialog dlg)
;;


let call_dialog self dlg =
  (* Call the dialog [dlg] from [self] as sub-dialog *)
  dlg # set_variable "nav_prev" (Dialog_value (Some (self :> dialog_type)));
  raise(Change_dialog dlg)
;;


let rec return_to_caller ?(n=1) self =
  (* Return from [self] to the calling dialog. The number [n] can be used
   * to go directly back to older antecedents.
   *)
  match self#dialog_variable "nav_prev" with
      Some dlg ->
	if n<=1 then (
	  dlg # set_next_page dlg#page_name;
	  raise(Change_dialog dlg)
	)
	else 
	  return_to_caller ~n:(n-1) self
    | None ->
	failwith "Cannot return to calling dialog - there is none!"
;;



class virtual std_behaviour =
object(self)

  method private virtual try_prepare_page : unit -> unit

  method private virtual try_handle : unit -> unit

  method private virtual check_auth : unit -> unit

  method private prepare_error_page error =
    let errid, detail = errid_of_error error in
    self # set_variable "runtime-error" (String_value errid);
    self # set_variable "runtime-error-detail" (String_value detail);
    self # set_next_page "runtime-error"

  method private handle_error error =
    self # prepare_error_page error;
    ()

  method private menu_handle() =
    match self#event with
	Button "nav-cancel" ->
	  return_to_caller self
      | Indexed_button("nav-back",n) ->
	  return_to_caller ~n:(int_of_string n) self
      | _ ->
	  ()

  method prepare_page() =
    try
      self # check_auth();
      self # try_prepare_page()
    with
      | Wd_types.Change_dialog _ as exn -> raise exn
      | Wd_types.Change_page _ as exn -> raise exn
      | error ->
	  self # prepare_error_page error

  method handle() =
    try
      self # check_auth();
      self # menu_handle();
      self # try_handle()
    with
      | Wd_types.Change_dialog _ as exn -> raise exn
      | Wd_types.Change_page _ as exn -> raise exn
      | error ->
	  self # handle_error error

  method private creds =
    Common.get_creds 
      (self # application # dtd)
      (try self # string_variable "password" with No_such_variable _ -> "")
end


(* Actually, a bug in WDialog, so we define a new output encoding: *)
(* (This bug is present in WDialog 2.0.1, but not in later versions!) *)

let subst rex f s =
  (* Copied from Wd_encoding *)
  let matches = try Pcre.exec_all ~rex s with Not_found -> [||] in
  let substitutes =
    Array.map (fun m ->
                 let (pos,pos') = Pcre.get_substring_ofs m 0 in
                 f pos (Pcre.get_substring m 0)) matches in
  let newlen = ref (String.length s) in
  Array.iteri
    (fun i m ->
       let sb = substitutes.(i) in
       let p0,p1 = Pcre.get_substring_ofs m 0 in
       newlen := !newlen - (p1-p0) + String.length sb
    )
    matches;
  let s' = String.create !newlen in
  let k = ref 0 in    (* position in s *)
  let k' = ref 0 in   (* position in s' *)
  Array.iteri
    (fun i m ->
       let sb = substitutes.(i) in
       let p0,p1 = Pcre.get_substring_ofs m 0 in
       String.blit s !k s' !k' (p0 - !k);
       k' := !k' + (p0 - !k);
       String.blit sb 0 s' !k' (String.length sb);
       k' := !k' + (String.length sb);
       k := p1;
    )
    matches;
  String.blit s !k s' !k' (String.length s - !k);
  assert(!k' + String.length s - !k = String.length s');
  s'
;;

let repeat n s =
  (* Copied from Wd_encoding *)
  (* Repeat the string [s] [n] times *)
  let l = String.length s in
  let s' = String.create (l * n) in
  for k = 0 to n-1 do
    String.blit s 0 s' (k*l) l
  done;
  s'
;;

let encode_as_pre_re = Pcre.regexp "[ \r\n\t]";;
  (* Copied from Wd_encoding *)
 
let encode_as_pre =
  let last_nl = ref 0 in
  subst
    encode_as_pre_re
    (fun pos ->
       function
           " "  -> "&nbsp;"
         | "\r" -> ""   (* simply drop *)
         | "\n" -> last_nl := pos+1  (* HERE WAS THE ERROR *) ; "<br>"
         | "\t" ->
             let spaces = 8 - (pos - !last_nl) mod 8 in
             last_nl := pos+1;   (* HERE WAS THE ERROR, TOO *)
             repeat spaces "&nbsp;"
         | _ ->
             assert false
    )
;;
